/**
 * Computes the SVD on a data matrix
 * that has the hodogram of an ellipse
 * @author: Joseph Jennings
 * @version: 2017.02.20
 */

#include "sep_params.h"
#include "sregc.h"
#include "hypercube_complex.h"
#include <Eigen/Dense>

#define DEG2RAD M_PI/180.0

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc,argv);
  std::string pvfile = pars.get_string("pvec","");
  std::string ufile  = pars.get_string("uout","");
  float theta = pars.get_float("theta",0.0);
  float sc1 = pars.get_float("scale1",1.0);
  float sc2 = pars.get_float("scale2",1.0);

  sregc fin("in");

  int n = fin.get_axis(1).n;

  std::vector<axis> faxes;
  faxes.push_back(axis(2*n,0.0,fin.get_axis(1).d));

  hypercube_complex *ellin = new hypercube_complex(fin.get_axes(),true);
  hypercube_float   *elflt = new hypercube_float(faxes,true);

  fin.read_all("in",ellin);

  /* Pack complex into a long float */
  for(int k = 0; k < n; ++k) {
    elflt->vals[2*k  ] = ellin->vals[k].real();
    elflt->vals[2*k+1] = ellin->vals[k].imag();
  }

  /* Build the nt x 2 data matrix */
  Eigen::MatrixXf D = Eigen::Map<Eigen::Matrix<float, Eigen::Dynamic, 2, Eigen::RowMajor> >(elflt->vals,n,2);

  //std::cerr << "D: " << std::endl << D << std::endl;

  /* Compute the SVD */
  Eigen::JacobiSVD<Eigen::MatrixXf> Dsvd(D, Eigen::ComputeThinU | Eigen::ComputeThinV);
  Eigen::MatrixXf S(2,2);
  S = Dsvd.singularValues().asDiagonal();
  Eigen::MatrixXf V = Dsvd.matrixV();
  Eigen::MatrixXf U = Dsvd.matrixU();

  /* Scaled singular vectors */
  Eigen::MatrixXf SV = S * V.transpose();

  /* Apply rotation operator to each polarization vector */
  Eigen::MatrixXf RV(2,2);
  RV(0,0) = cos(theta*DEG2RAD)*V(0,0) - sin(theta*DEG2RAD)*V(1,0);
  RV(1,0) = sin(theta*DEG2RAD)*V(0,0) + cos(theta*DEG2RAD)*V(1,0);
  RV(0,1) = cos(theta*DEG2RAD)*V(0,1) - sin(theta*DEG2RAD)*V(1,1);
  RV(1,1) = sin(theta*DEG2RAD)*V(0,1) + cos(theta*DEG2RAD)*V(1,1);

  /* Scale singular values */
  Eigen::MatrixXf SC(2,2);
  SC(0,0) = S(0,0)*sc1; SC(1,1) = S(1,1)*sc2;
  SC(0,1) = 0.0; SC(1,0) = 0.0;

  /* Build filtered data */
  Eigen::MatrixXf Drsc = U*SC*RV.transpose();
  //Eigen::MatrixXf Drsc = U*S*V.transpose();

//  std::cerr << "V: " << std::endl << V << std::endl;
//  std::cerr << "S: " << std::endl << S << std::endl << std::endl;
//
//  std::cerr << "RV: " << std::endl << RV << std::endl;
//  std::cerr << "SC: " << std::endl << SC << std::endl;
//  std::cerr << "Drsc: " << std::endl << Drsc << std::endl;

  /* Write out data */
  hypercube_float *dout    = new hypercube_float(faxes,Drsc.data());
  hypercube_complex *dcout = new hypercube_complex(fin.get_axes(),true);
  for(int i = 0; i < n; ++i) {
    dcout->vals[i].real(dout->vals[0*n + i]);
    dcout->vals[i].imag(dout->vals[1*n + i]);
  }
  sregc wrtdat(dcout->get_axes());
  wrtdat.write_description("out");
  wrtdat.write_all("out",dcout);

  /* Write out SV */
  if(!pvfile.empty()) {
    std::vector<axis> paxis;
    paxis.push_back(axis(3,0.0,1.0));
    hypercube_complex *pvecs = new hypercube_complex(paxis,true);
    /* Build output polarization vector */
    pvecs->vals[0].real(0.0);  pvecs->vals[0].imag(0.0);
    pvecs->vals[1].real(SV.data()[0]); pvecs->vals[1].imag(SV.data()[2]);
    pvecs->vals[2].real(SV.data()[1]); pvecs->vals[2].imag(SV.data()[3]);

    sregc pout(pvecs->get_axes());
    pout.write_description("pvec");
    pout.write_all("pvec",pvecs);
  }

  /* Write out U */
  if(!ufile.empty()) {
    hypercube_float *uf = new hypercube_float(faxes,U.data());
    hypercube_complex *uout = new hypercube_complex(fin.get_axes(),true);
    for(int i = 0; i < n; ++i) {
      uout->vals[i].real(uf->vals[0*n + i]);
      uout->vals[i].imag(uf->vals[1*n + i]);
    }
    sregc wrtu(uout->get_axes());
    wrtu.write_description("uout");
    wrtu.write_all("uout",uout);
  }

  delete elflt; delete ellin;
  delete dout; delete dcout;
}
