/**
 * Performs polarization analysis on
 * multicomponent shotgathers via SVD
 * @author: Joseph Jennings
 * @version: 2017.02.21
 */

#include "sep_params.h"
#include "hypercube_complex.h"
#include "sregc.h"
#include <Eigen/Dense>

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc,argv);
  float thresh = pars.get_float("thresh",-999.0);

  /* Read in the three component data */
  sregf mcin = sregf("in");

  int nt = mcin.get_axis(1).n;
  int nr = mcin.get_axis(2).n;
  int ns = mcin.get_axis(3).n;
  int nc = mcin.get_axis(4).n;
  float dt = mcin.get_axis(1).d;

  float twind = pars.get_float("twind",(nt)*dt);
  int nwind = static_cast<int>(twind/dt + 0.5);

  if(nc != 3) seperr("Error: can only perform polarization analysis of three "
      "component data (Pressure, Z and X components) for now");

  std::vector<axis> ocaxes,oaxes,at;
  /* Single component axes */
  ocaxes.push_back(mcin.get_axis(1));
  ocaxes.push_back(mcin.get_axis(2));
  ocaxes.push_back(mcin.get_axis(3));
  /* Output polarization axes */
  oaxes.push_back(mcin.get_axis(1));
  oaxes.push_back(mcin.get_axis(2));
  // 5 = 0,0 + polarization vectors + repeated singular values
  // 2*nwind is for U and the hodogram
  int np = 7 + 2*nwind;
  oaxes.push_back(axis(np,0.0,1.0));
  oaxes.push_back(mcin.get_axis(3));
  /* Single trace axis */
  at.push_back(mcin.get_axis(1));

  hypercube_float *dat  = new hypercube_float(mcin.get_axes(),true);
  hypercube_float *zcmp = new hypercube_float(ocaxes,true);
  hypercube_float *xcmp = new hypercube_float(ocaxes,true);
  hypercube_float *ztr  = new hypercube_float(at,true);
  hypercube_float *xtr  = new hypercube_float(at,true);
  hypercube_complex *pol  = new hypercube_complex(oaxes,true);

  mcin.read_all("in",dat);

  /* Get the z component */
  memcpy(zcmp->vals,&dat->vals[(1)*nt*nr*ns],sizeof(float)*(1)*nt*nr*ns);
  /* Get the x component */
  memcpy(xcmp->vals,&dat->vals[(2)*nt*nr*ns],sizeof(float)*(1)*nt*nr*ns);

  /* Eigenmatrices */
  Eigen::MatrixXf D(nwind,2);
  Eigen::MatrixXf U(nwind,2);
  Eigen::Matrix2f V;
  Eigen::Matrix2f S;

  /* Loop over each trace in the dataset */
  for(int is = 0; is < ns; ++is) {
    for(int ir = 0; ir < nr; ++ir) {
      /* Get the two components */
      memcpy(ztr->vals,&zcmp->vals[nr*nt*(is) + nt*(ir)],sizeof(float)*nt);
      memcpy(xtr->vals,&xcmp->vals[nr*nt*(is) + nt*(ir)],sizeof(float)*nt);
      for(int it = 0; it < nt; ++it) {
        /* Build the data matrix */
        for(int k = 0; k < nwind; ++k) {
          if(it+k < nt) {
            D(k,0) = ztr->vals[it+k];
            D(k,1) = xtr->vals[it+k];
            /* Hodogram */
            pol->vals[is*np*nt*nr + (k+7+nwind)*nr*nt + ir*nt + it] = std::complex<float>(D(k,0),D(k,1));
          }
        }
        /* Compute the SVD */
        Eigen::JacobiSVD<Eigen::MatrixXf> Dsvd(D, Eigen::ComputeThinU | Eigen::ComputeThinV);
        V = Dsvd.matrixV(); S = Dsvd.singularValues().asDiagonal(); U = Dsvd.matrixU();
        /* Save polarization parameters */
        pol->vals[is*np*nt*nr + (0)*nr*nt + ir*nt + it] = std::complex<float>(0.0,0.0);
        if(thresh != -999.0) {
          if(S(0,0) < thresh ||  S(1,1) < thresh) {
            pol->vals[is*np*nt*nr + (1)*nr*nt + ir*nt + it] = 0.0;
            pol->vals[is*np*nt*nr + (2)*nr*nt + ir*nt + it] = 0.0;
          }
          else {
            pol->vals[is*np*nt*nr + (1)*nr*nt + ir*nt + it] = std::complex<float>(V(0,0),V(1,0));
            pol->vals[is*np*nt*nr + (2)*nr*nt + ir*nt + it] = std::complex<float>(V(0,1),V(1,1));
          }
        } else {
          pol->vals[is*np*nt*nr + (1)*nr*nt + ir*nt + it] = std::complex<float>(V(0,0),V(1,0));
          pol->vals[is*np*nt*nr + (2)*nr*nt + ir*nt + it] = std::complex<float>(V(0,1),V(1,1));
        }
        pol->vals[is*np*nt*nr + (3)*nr*nt + ir*nt + it] = std::complex<float>(0.0,0.0);
        pol->vals[is*np*nt*nr + (4)*nr*nt + ir*nt + it] = std::complex<float>(S(0,0),0.0);
        pol->vals[is*np*nt*nr + (5)*nr*nt + ir*nt + it] = std::complex<float>(0.0,0.0);
        pol->vals[is*np*nt*nr + (6)*nr*nt + ir*nt + it] = std::complex<float>(S(1,1),0.0);
        /* Write out U */
        for(int iu = 0; iu < nwind; ++iu) {
          pol->vals[is*np*nt*nr + (iu+7)*nr*nt + ir*nt + it] = std::complex<float>(U(iu,0),U(iu,1));
        }
      }
    }
  }

  sregc pout(pol->get_axes());
  pout.write_description("out");
  pout.write_all("out",pol);
  putch("twind","f",&twind);
  putch("nwind","d",&nwind);

  delete dat; delete zcmp; delete xcmp;
  delete ztr; delete xtr; delete pol;
}
