/**
 * Filters the blended data via inner-products
 * of the polarization vectors computed
 * via SVD.
 * @author: Joseph Jennings
 * @version: 2017.02.21
 */

#include "sep_params.h"
#include "hypercube_float.h"
#include "sregc.h"
#include <Eigen/Dense>

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc,argv);
  float exp = pars.get_float("exp",1.0);
  std::string filtin = pars.get_string("filt","");
  std::string wtout  = pars.get_string("wt","");
  if(filtin.empty()) {
    seperr("Error: Need a user-provided filter as input. Please compute with PolAnalysisSVD.\n");
  }

  /* Read in the three component blended data */
  sregf mcin = sregf("in");
  sregc flin = sregc("filt");

  int nt = mcin.get_axis(1).n;
  int nr = mcin.get_axis(2).n;
  int nbs = mcin.get_axis(3).n;
  int ns  = nbs*2; // Assumes two shots per blended shot
  int np = flin.get_axis(3).n;

  /* Get windows used for computing the filter */
  int nwind = 0;
  auxpar("nwind","d",&nwind,filtin.c_str());

  std::vector<axis> oaxes,ocaxes,faxes,waxes,at;
  /* Output axes */
  oaxes.push_back(mcin.get_axis(1));
  oaxes.push_back(mcin.get_axis(2));
  /* Now has twice as many shots */
  oaxes.push_back(axis(ns,0.0,1.0));
  oaxes.push_back(mcin.get_axis(4));
  /* Single component axes */
  ocaxes.push_back(mcin.get_axis(1));
  ocaxes.push_back(mcin.get_axis(2));
  ocaxes.push_back(mcin.get_axis(3));
  /* Weight function axis */
  waxes.push_back(mcin.get_axis(1));
  waxes.push_back(mcin.get_axis(2));
  waxes.push_back(mcin.get_axis(3));
  waxes.push_back(axis(ns,0.0,1.0));
  /* Single trace axis */
  at.push_back(mcin.get_axis(1));

  hypercube_float *bld  = new hypercube_float(mcin.get_axes(),true);
  hypercube_float *dbld = new hypercube_float(oaxes,true);
  hypercube_float *zcmp = new hypercube_float(ocaxes,true);
  hypercube_float *xcmp = new hypercube_float(ocaxes,true);
  hypercube_float *pcmp = new hypercube_float(ocaxes,true);
  hypercube_float *wt   = new hypercube_float(waxes,true);
  hypercube_float *ztr  = new hypercube_float(at,true);
  hypercube_float *xtr  = new hypercube_float(at,true);
  hypercube_float *ptr  = new hypercube_float(at,true);
  hypercube_complex *pf = new hypercube_complex(flin.get_axes(),true);

  mcin.read_all("in",bld);
  flin.read_all("filt",pf);

  /* Get the pressure */
  memcpy(pcmp->vals,&bld->vals[(0)*nt*nr*nbs],sizeof(float)*(1)*nt*nr*nbs);
  /* Get the z component */
  memcpy(zcmp->vals,&bld->vals[(1)*nt*nr*nbs],sizeof(float)*(1)*nt*nr*nbs);
  /* Get the x component */
  memcpy(xcmp->vals,&bld->vals[(2)*nt*nr*nbs],sizeof(float)*(1)*nt*nr*nbs);

  /* Eigenmatrices */
  Eigen::MatrixXf D(nwind,2);
  Eigen::MatrixXf U(nwind,2);
  Eigen::Matrix2f V;
  Eigen::Matrix2f S;

  /* Loop over each trace in the dataset */
  for(int is = 0; is < nbs; ++is) {
    for(int ir = 0; ir < nr; ++ir) {
      /* Get the three components */
      memcpy(ztr->vals,&zcmp->vals[nr*nt*(is) + nt*(ir)],sizeof(float)*nt);
      memcpy(xtr->vals,&xcmp->vals[nr*nt*(is) + nt*(ir)],sizeof(float)*nt);
      memcpy(ptr->vals,&pcmp->vals[nr*nt*(is) + nt*(ir)],sizeof(float)*nt);
      for(int it = 0; it < nt; ++it) {
        /* Build the data matrix */
        for(int k = 0; k < nwind; ++k) {
          if(it+k < nt) {
            D(k,0) = ztr->vals[it+k];
            D(k,1) = xtr->vals[it+k];
          }
        }
        /* Compute the SVD */
        Eigen::JacobiSVD<Eigen::MatrixXf> Dsvd(D, Eigen::ComputeThinU | Eigen::ComputeThinV);
        V = Dsvd.matrixV(); S = Dsvd.singularValues().asDiagonal(); U = Dsvd.matrixU();
        /* Get principal vectors from ind. modeled data */
        float f1z = pf->vals[(is  )*np*nt*nr + (1)*nr*nt + ir*nt + it].real();
        float f1x = pf->vals[(is  )*np*nt*nr + (1)*nr*nt + ir*nt + it].imag();
        float f2z = pf->vals[(is+1)*np*nt*nr + (1)*nr*nt + ir*nt + it].real();
        float f2x = pf->vals[(is+1)*np*nt*nr + (1)*nr*nt + ir*nt + it].imag();
        /* Compute inner products */
        //TODO: Smooth this weight
        float wt1 = fabs(f1z*V(0,0) + f1x*V(1,0));
        float wt2 = fabs(f2z*V(0,0) + f2x*V(1,0));
        if(!wtout.empty()) {
          wt->vals[0*nbs*nr*nt + (is)*nr*nt + (ir)*nt + it] = wt1;
          wt->vals[1*nbs*nr*nt + (is)*nr*nt + (ir)*nt + it] = wt2;
        }
        /* Apply filter to pressure */
        dbld->vals[0*nt*nr*ns + (is  )*nr*nt + (ir)*nt + it] = pow(wt1,exp)*ptr->vals[it];
        dbld->vals[0*nt*nr*ns + (is+1)*nr*nt + (ir)*nt + it] = pow(wt2,exp)*ptr->vals[it];
        /* Apply filter to Z component */
        dbld->vals[1*nt*nr*ns + (is  )*nr*nt + (ir)*nt + it] = pow(wt1,exp)*ztr->vals[it];
        dbld->vals[1*nt*nr*ns + (is+1)*nr*nt + (ir)*nt + it] = pow(wt2,exp)*ztr->vals[it];
        /* Apply filter to X component */
        dbld->vals[2*nt*nr*ns + (is  )*nr*nt + (ir)*nt + it] = pow(wt1,exp)*xtr->vals[it];
        dbld->vals[2*nt*nr*ns + (is+1)*nr*nt + (ir)*nt + it] = pow(wt2,exp)*xtr->vals[it];
      }
    }
  }

  /* Write out deblended */
  sregf dbout(dbld->get_axes());
  dbout.write_description("out");
  dbout.write_all("out",dbld);

  if(!wtout.empty()) {
    sregf wtreg(wt->get_axes());
    wtreg.write_description("wt");
    wtreg.write_all("wt",wt);
  }

  /* Free memory */
  delete bld; delete dbld; delete pf; delete wt;
  delete zcmp; delete xcmp; delete pcmp;
  delete ztr; delete xtr; delete ptr;
}
