/*
 * acfdprop2d.cpp
 *
 *  Created on: Nov 26, 2016
 *      Author: joe
 */

#include "acfdprop2d.h"
#include "fdcoefnb.h"
#include "fdcoefb.h"

acfdprop2d::acfdprop2d(std::vector<axis> uaxis, hypercube_float *vel, laplacian* lapop, float fmax) {
  /* Get temporal and spatial grid parameters */
  _nz = uaxis[0].n; _nx = uaxis[1].n; _nt = uaxis[2].n;
  _dz = uaxis[0].d; _dx = uaxis[1].d; _dt = uaxis[2].d;
  std::vector<axis> azx; azx.push_back(uaxis[0]); azx.push_back(uaxis[1]);

  _lapop = lapop;

  /* Get velocity and frequency info */
  _fmax = fmax; _vmax = vel->my_max(); _vmin = find_vmin(vel);

  /* Check spatial samplings */
  if(!check_dxdz(lapop)) {
    seperr("Spatial samplings in laplacian operator are not equivalent to "
        "ACFDprop2d spatial samplings. Please check again.\n");
  }

  /* Check for stability */
  if(!check_cfl(_vmax)) {
    fprintf(stderr,"CFL is: %f\n",get_cfl());
    seperr("ACFDprop2d: Error, please change your samplings. Your propagation will be"
        " unstable.\n");
  }

  /* Check for dispersion */
  if(!check_ptsperlength(_vmin, _fmax)) {
    fprintf(stderr,"PPL is: %f\n",get_ptsperlength());
    seperr("ACFDprop2d: Error, please change your samplings. Your propagation will be"
        " dispersive.\n");
  }

  /* Precompute velocity dt^2 coeffiecient */
  _v2dt2 = (hypercube_float*)vel->clone_vec();
  _v2dt2->mult(_v2dt2);
  _v2dt2->scale(_dt * _dt);

  /* Create FD coef operator */
  _coefop = new fdcoefnb(azx, _lapop, _v2dt2);

  /* Allocate for memory for wavefield slices*/
  _slc = new hypercube_float(azx, true);
  _dlc = new hypercube_float(azx, true);
  _w = new hypercube_float(vel->get_axes(), true);
  _w->zero();

  _wuout.tagit = "";
}

acfdprop2d::acfdprop2d(std::vector<axis> uaxis, hypercube_float *vel, hypercube_float *wgt, hypercube_float *vwt,
    std::vector<int> nbs, laplacian* lapop, float fmax) {
  /* Get temporal and spatial grid parameters */
  _nz = uaxis[0].n; _nx = uaxis[1].n; _nt = uaxis[2].n;
  _dz = uaxis[0].d; _dx = uaxis[1].d; _dt = uaxis[2].d;
  std::vector<axis> azx; azx.push_back(uaxis[0]); azx.push_back(uaxis[1]);

  _lapop = lapop;

  /* Get velocity and frequency info */
  _fmax = fmax; _vmax = vel->my_max(); _vmin = find_vmin(vel);

  /* Check spatial samplings */
  if(!check_dxdz(lapop)) {
    seperr("Spatial samplings in laplacian operator are not equivalent to "
        "ACFDprop2d spatial samplings. Please check again.\n");
  }

  /* Check for stability */
  if(!check_cfl(_vmax)) {
    fprintf(stderr,"CFL is: %f\n",get_cfl());
    seperr("ACFDprop2d: Error, please change your samplings. Your propagation will be"
        " unstable.\n");
  }


  /* Check for dispersion */
  if(!check_ptsperlength(_vmin, _fmax)) {
    fprintf(stderr,"PPL is: %f\n",get_ptsperlength());
    seperr("ACFDprop2d: Error, please change your samplings. Your propagation will be"
        " dispersive.\n");
  }

  /* Precompute velocity dt^2 coeffiecient */
  _v2dt2 = (hypercube_float*)vel->clone_vec();
  _v2dt2->mult(_v2dt2);
  _v2dt2->scale(_dt * _dt);

  /* Create FD coef operator */
  _coefop = new fdcoefb(uaxis, _lapop, _v2dt2, wgt, vwt, nbs);
  _w = wgt;

  /* Allocate for memory for wavefield slices*/
  _slc = new hypercube_float(azx, true);
  _dlc = new hypercube_float(azx, true);

  _wuout.tagit = "";
}

bool acfdprop2d::forward(bool add, my_vector *mod, my_vector *dat) {
  if(!add && dat) dat->zero();

  /* Initialize temporary slices */
  _slc->zero(); _dlc->zero();

  if(!mod->check_match(dat)) {
    seperr("Input source wavelet and output wavefield must have same axes.\n");
  }

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  /*  Copy for it == 0 */
  memcpy(d->vals,m->vals,sizeof(float)*_nx*_nz);

  memcpy(_dlc->vals,d->vals,sizeof(float)*_nx*_nz);
  _coefop->forward(false,_dlc,_slc);

  /* Copy for it = 1 */
  for(int ix = 0; ix < _nx; ++ix) {
    for(int iz = 0; iz < _nz; ++iz) {
      d->vals[_nx*_nz + ix*+_nz + iz] += m->vals[_nx*_nz + ix*_nz + iz] + _slc->vals[ix*_nz + iz];
    }
  }

  int t0, t1, t2, tx;
  for(int it = 2; it < _nt; ++it) {
    fprintf(stderr,"\b\b\b\b\b\b\bit=%d",it);
    /* Compute time shift */
    t0 = (it-0)*_nx*_nz;
    t1 = (it-1)*_nx*_nz;
    t2 = (it-2)*_nx*_nz;
    /* Apply laplacian */
    memcpy(_dlc->vals,&d->vals[t1],sizeof(float)*_nx*_nz);
    _coefop->forward(false,_dlc,_slc);
    for(int ix = 0; ix < _nx; ++ix) {
      /* Compute x shift */
      tx = ix*_nz;
      for(int iz = 0; iz < _nz; ++iz) {
        int it0 = t0 + tx + iz;
        int it2 = t2 + tx + iz;
        int vidx =     tx + iz;
        d->vals[it0] += m->vals[it0] + _slc->vals[vidx] + (_w->vals[vidx]-1)*d->vals[it2];
      }
    }
  }
  fprintf(stderr,"\n");
  return true;
}

bool acfdprop2d::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(!add && mod) mod->zero();

  /* Initialize temporary slices */
  _slc->zero(); _dlc->zero();

  if(!mod->check_match(dat)) {
    seperr("Input source wavefield and output wavelet must have same axes.\n");
  }

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  /* Copy for it = nt-1*/
  memcpy(&m->vals[(_nt-1)*_nx*_nz],&d->vals[(_nt-1)*_nx*_nz],sizeof(float)*_nx*_nz);

  memcpy(_dlc->vals,&m->vals[(_nt-1)*_nx*_nz],sizeof(float)*_nx*_nz);
  _coefop->adjoint(false,_slc,_dlc);

  /* nt-2 initial condition */
  for(int ix = 0; ix < _nx; ++ix) {
    for(int iz = 0; iz < _nz; ++iz) {
      m->vals[(_nt-2)*_nx*_nz + ix*+_nz + iz] += d->vals[(_nt-2)*_nx*_nz + ix*_nz + iz] + _slc->vals[ix*_nz + iz];
    }
  }

  int t0, t1, t2, tx;
  for(int it = _nt-3; it >= 0; --it) {
    fprintf(stderr,"\b\b\b\b\b\b\bit=%d",it);
    fprintf(stderr," ");
    /* Compute time shift */
    t0 = (it+0)*_nx*_nz;
    t1 = (it+1)*_nx*_nz;
    t2 = (it+2)*_nx*_nz;
    /* Apply laplacian */
    memcpy(_dlc->vals,&m->vals[t1],sizeof(float)*_nx*_nz);
    _coefop->adjoint(false,_slc,_dlc);
    for(int ix = 0; ix < _nx; ++ix) {
      /* Compute x shift */
      tx = ix*_nz;
      for(int iz = 0; iz < _nz; ++iz) {
        int it0 = t0 + tx + iz;
        int it2 = t2 + tx + iz;
        int vidx =     tx + iz;
        m->vals[it0] += d->vals[it0] + _slc->vals[vidx] + (_w->vals[vidx]-1)*m->vals[it2];
      }
    }
  }
  return true;
}

bool acfdprop2d::check_cfl(float vmax) {
  _mycfl = (_dt * vmax)/(std::min(_dx,_dz));
  if(_mycfl < _CFL) {
    return true;
  } else {
    return false;
  }
}

bool acfdprop2d::check_ptsperlength(float vmin, float fmax) {
  _ppl = vmin/(fmax*std::max(_dx,_dz));
  if(_lapop->get_myorder() == 4) {
    if(_ppl < 4) {
      return false;
    } else {
      return true;
    }
  }
  else if(_lapop->get_myorder() == 10) {
    //TODO: Condition from Dablain paper?
    //      return true for now
    return true;
  }
  else{
    seperr("ACFDprop2d has not been implemented for stencil orders besides 4 and 10.\n");
    return false;
  }
}

float acfdprop2d::find_vmin(hypercube_float *vel) {

  int shift = _lapop->get_myorder()/2;
  float vmin = vel->vals[shift];

  for(int ix = shift; ix < _nx - shift; ++ix) {
    for(int iz = shift; iz < _nz - shift; ++iz) {
      if(vmin < vel->vals[ix*_nz + iz])
        vmin = vel->vals[ix*_nz + iz];
    }
  }

  return vmin;
}

bool acfdprop2d::check_dxdz(laplacian *lapop) {
  float idx2 = 1/(_dx*_dx); float idz2 = 1/(_dz*_dz);
  if(lapop->_idx2 != idx2 && lapop->_idz2 != idz2) {
    return false;
  } else {
    return true;
  }
}

float acfdprop2d::get_cfl() {
  check_cfl(_vmax);
  return _mycfl;
}

float acfdprop2d::get_ptsperlength() {
  check_ptsperlength(_vmin, _fmax);
  return _ppl;
}
