/*
 * afd2dfst.cpp
 *
 *  Created on: Jan 3, 2017
 *      Author: joe
 */

#include "afd2dfst.h"

afd2dfst::afd2dfst(std::vector<axis> uaxes, hypercube_float *vel, laplacian* lapop, float dtu, float fmax, int bz, int bx, float alpha) {
  init(uaxes, vel, lapop, dtu, fmax, bz, bx, alpha);
}

afd2dfst::afd2dfst(std::vector<axis> uaxes, hypercube_float *vel, laplacian* lapop, float dtu, float fmax) {
  init(uaxes, vel, lapop, dtu, fmax, 50, 50, 0.97);
}

void afd2dfst::init(std::vector<axis> uaxes, hypercube_float *vel, laplacian* lapop, float dtu, float fmax, int bz, int bx, float alpha) {
  /* Get temporal and spatial grid parameters */
  _nz = uaxes[0].n; _nx = uaxes[1].n;  _nts = uaxes[2].n;
  _dz = uaxes[0].d; _dx = uaxes[1].d;  _dts = uaxes[2].d;
  _dtu = dtu;
  _onestp = _nx*_nz; _sample = static_cast<int>(_dts/_dtu);
  std::vector<axis> azx; azx.push_back(uaxes[0]); azx.push_back(uaxes[1]);

  _lapop = lapop;

  /* Get velocity and frequency info */
  _fmax = fmax; _vmax = vel->my_max(); _vmin = find_vmin(vel);

  /* Check spatial samplings */
  if(!check_dxdz(lapop)) {
    seperr("Spatial samplings in laplacian operator are not equivalent to "
        "ACFDprop2d spatial samplings. Please check again.\n");
  }

  /* Check for stability */
  if(!check_cfl(_vmax,lapop)) {
    fprintf(stderr,"myCFL: %f lapCFL: %f\n",get_cfl().first,get_cfl().second);
    seperr("ACFDprop2d: Error, please change your samplings. Your propagation will be"
        " unstable.\n");
  }

  /* Check for dispersion */
  if(!check_ptsperlength(_vmin,_fmax,lapop)) {
    fprintf(stderr,"PPL is: %f\n",get_ptsperlength());
    seperr("ACFDprop2d: Error, please change your samplings. Your propagation will be"
        " dispersive.\n");
  }

  /* Precompute velocity dt^2 coeffiecient */
  _v2dt2 = (hypercube_float*)vel->clone_vec();
  _v2dt2->mult(_v2dt2);
  _v2dt2->scale(_dtu * _dtu);

  /* Build taper */
  _spg = new tapercos(azx,bz,bx,_lapop->get_myorder(),alpha);

  //FIXME: Probably not the best way of doing this
  // might be better to pass four vectors by reference
  /* Interpolation coefficients forward */
  _ls1 = interp_coeff(_sample).first;
  _ls2 = interp_coeff(_sample).second;
  /* Interpolation coefficients adjoint */
  _lr1 = interp_coefa(_sample).first;
  _lr2 = interp_coefa(_sample).second;

  /* Allocate memory for wavefield slices*/
  _pre   = new hypercube_float(azx, true);
  _cur   = new hypercube_float(azx, true);
  _nex   = new hypercube_float(azx, true);
  _sou1  = new hypercube_float(azx, true);
  _sou2  = new hypercube_float(azx, true);
  _tsou1 = new hypercube_float(azx, true);
  _tsou2 = new hypercube_float(azx, true);
  _lap   = new hypercube_float(azx, true);
  //TODO: Do not allocate tmp
  _tmp   = new hypercube_float(azx, true);

}

void afd2dfst::set_bndpars(int bz, int bx, float alpha) {
  std::vector<axis> azx;
  azx.push_back(axis(_nz,0.0,_dz)); azx.push_back(axis(_nx,0.0,_dx));
  _spg = new tapercos(azx,bz,bx,_lapop->get_myorder(),alpha);
}

std::pair<std::vector<double>,std::vector<double>> afd2dfst::interp_coeff(int sample) {
  std::vector<double> ls1, ls2;
  for(int is = 0; is < sample; ++is) {
    ls1.push_back(static_cast<double>((sample-is))/static_cast<double>(sample));
    ls2.push_back(1.0-ls1[is]);
    ls1[is] /= sample; ls2[is] /= sample;
  }
  return std::pair<std::vector<double>,std::vector<double>>(ls1,ls2);
}

std::pair<std::vector<double>,std::vector<double>> afd2dfst::interp_coefa(int sample) {
  std::vector<double> lr1, lr2;
  for(int is = 0; is < sample; ++is) {
    lr1.push_back(static_cast<double>((sample-is-1))/static_cast<double>(sample));
    lr2.push_back(1.0-lr1[is]);
    lr1[is] /= sample; lr2[is] /= sample;
  }
  return std::pair<std::vector<double>,std::vector<double>>(lr1,lr2);
}

bool afd2dfst::forward(bool add, my_vector *mod, my_vector *dat) {
  if(add && dat) dat->zero();

  hypercube_float* m = (hypercube_float*)mod;
  hypercube_float* d = (hypercube_float*)dat;

  /* Initialize temporary slices */
  _pre->zero(); _cur->zero(); _nex->zero(); _sou1->zero(); _tmp->zero();
  _sou2->zero(); _tsou1->zero(); _tsou2->zero();

  /* it = -1 (u'(0) = 0) */
  memcpy(d->vals,_pre->vals,sizeof(float)*_onestp);
  /* it = 0 (u(0) = 0) */
  memcpy(&d->vals[_onestp],_pre->vals,sizeof(float)*_onestp);

  /* Time loop */
  //TODO: Write a unit test to compare this wavefield
  //      to a wavefield interpolated outside
  for(int it = 1; it < _nts-1; ++it) {
    /* Extract source */
    memcpy(_sou1->vals,&m->vals[(it-1)*_onestp],sizeof(float)*_onestp);
    memcpy(_sou2->vals,&m->vals[(it-0)*_onestp],sizeof(float)*_onestp);
    /* Interpolate source and add */
    for(int is = 0; is < _sample; ++is) {
      /* Forward march */
      fmarch(_pre,_cur,_nex);
      /* Forward interpolate source */
      _nex->scale_add(1.0,_sou1,_ls1[is]);
      _nex->scale_add(1.0,_sou2,_ls2[is]);
      /* Apply taper */
      _spg->apply(_cur,_nex);
      /* Adjoint interpolation */
      _tsou1->scale_add(1.0,_nex,_ls1[is]);
      _tsou2->scale_add(1.0,_nex,_ls2[is]);
      /* Update time step */
      _tmp = _pre; _pre = _cur; _cur = _nex; _nex = _tmp;
    }
    /* Save to wavefield */
    memcpy(&d->vals[(it-1)*_onestp],_tsou1->vals,sizeof(float)*_onestp);
    memcpy(&d->vals[(it-0)*_onestp],_tsou2->vals,sizeof(float)*_onestp);
    _tsou1->set(_tsou2->vals);
    _tsou2->zero();
  }
  return true;
}

bool afd2dfst::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(add && mod) mod->zero();

  hypercube_float* m = (hypercube_float*)mod;
  hypercube_float* d = (hypercube_float*)dat;

  /* Initialize temporary slices */
  _pre->zero(); _cur->zero(); _nex->zero(); _sou1->zero(); _tmp->zero();
  _sou2->zero(); _tsou1->zero(); _tsou2->zero();

  /* it = nt (u'(nt-1) = 0) */
  memcpy(&m->vals[(_nts-1)*_onestp],_nex->vals,sizeof(float)*_onestp);
  /* it = nt-1 (u(nt-1) = ) */
  memcpy(&m->vals[(_nts-2)*_onestp],_nex->vals,sizeof(float)*_onestp);

  /* Time loop */
  for(int it = _nts-2; it >= 1; --it) {
    /* Extract source */
    memcpy(_sou1->vals,&d->vals[(it-0)*_onestp],sizeof(float)*_onestp);
    memcpy(_sou2->vals,&d->vals[(it-1)*_onestp],sizeof(float)*_onestp);
    /* Interpolate source and add */
    for(int is = 0; is < _sample; ++is) {
      /* Backward march */
      bmarch(_pre,_cur,_nex);
      /* Forward interpolate source */
      _pre->scale_add(1.0,_sou1,_lr1[is]);
      _pre->scale_add(1.0,_sou2,_lr2[is]);
      /* Absorbing boundary */
      _spg->apply(_cur,_pre);
      /* Adjoint interpolate source */
      _tsou1->scale_add(1.0,_pre,_lr1[is]);
      _tsou2->scale_add(1.0,_pre,_lr2[is]);
      /* Update time step */
      _tmp = _nex; _nex = _cur; _cur = _pre; _pre = _tmp;
    }
    /* Save to source */
    memcpy(&m->vals[(it-0)*_onestp],_tsou1->vals,sizeof(float)*_onestp);
    memcpy(&m->vals[(it-1)*_onestp],_tsou2->vals,sizeof(float)*_onestp);
    _tsou1->set(_tsou2->vals);
    _tsou2->zero();
  }
  return true;
}

void afd2dfst::fmarch(hypercube_float *pre, hypercube_float *cur, hypercube_float *nex) {
  _lapop->forward(false,cur,_lap);
  for(int k = 0; k < _onestp; ++k) {
    nex->vals[k] = _v2dt2->vals[k]*_lap->vals[k] + 2*cur->vals[k] - pre->vals[k];
  }
}

void afd2dfst::bmarch(hypercube_float *pre, hypercube_float *cur, hypercube_float *nex) {
  /* Apply diagonal */
  for(int k = 0; k < _onestp; ++k) {
    _lap->vals[k] = _v2dt2->vals[k]*cur->vals[k];
  }
  /* Adjoint of laplacian */
  _lapop->adjoint(false,pre,_lap);
  for(int k = 0; k < _onestp; ++k) {
    pre->vals[k] += 2*cur->vals[k] - nex->vals[k];
  }
}

bool afd2dfst::check_cfl(float vmax, laplacian *lapop) {
  /* Compute condition based on laplacian odd coefficients */
  float sum = 0.0;
  std::vector<double> coefs = lapop->get_mycoefs();
  for(int i = 1; i < (int)coefs.size()/2 + 1; ++i) { sum += coefs[2*i-1]; }
  sum *= 2; _CFL = sqrt(1/sum);
  /* Compute my CFL */
  _mycfl = (_dtu * vmax)/(std::min(_dx,_dz));
  if(_mycfl < _CFL) {
    return true;
  } else {
    return false;
  }
}

bool afd2dfst::check_ptsperlength(float vmin, float fmax, laplacian *lapop) {
  _ppl = vmin/(fmax*std::max(_dx,_dz));
  if(lapop->get_myorder() == 4) {
    if(_ppl < 4) {
      return false;
    } else {
      return true;
    }
  }
  //TODO: How can this be derived?
  else if(_lapop->get_myorder() == 10) {
    if(_ppl < 3) {
      return false;
    } else {
      return true;
    }
  }
  else{
    seperr("afd2dfst has not been implemented for stencil orders besides 4 and 10.\n");
    return false;
  }
}

float afd2dfst::find_vmin(hypercube_float *vel) {

  int shift = _lapop->get_myorder()/2;
  float vmin = vel->vals[(shift)*_nz + shift];

  for(int ix = shift; ix < _nx - shift; ++ix) {
    for(int iz = shift; iz < _nz - shift; ++iz) {
      if(vmin > vel->vals[ix*_nz + iz]) {
        vmin = vel->vals[ix*_nz + iz];
      }
    }
  }

  return vmin;
}

bool afd2dfst::check_dxdz(laplacian *lapop) {
  float idx2 = 1/(_dx*_dx); float idz2 = 1/(_dz*_dz);
  if(lapop->_idx2 != idx2 && lapop->_idz2 != idz2) {
    return false;
  } else {
    return true;
  }
}

std::pair<float,float> afd2dfst::get_cfl() {
  return std::pair<float,float>(_mycfl,_CFL);
}

float afd2dfst::get_ptsperlength() {
  return _ppl;
}
