/*
 * afd3d.cpp
 *
 *  Created on: Jan 9, 2017
 *      Author: joe
 */

#include "afd3d.h"
#include <algorithm>

afd3d::afd3d(std::vector<axis> uaxis, hypercube_float *vel, laplacian3d* lapop, float fmax) {
  /* Get temporal and spatial grid parameters */
  _nz = uaxis[0].n; _nx = uaxis[1].n; _ny = uaxis[2].n; _ntu = uaxis[3].n;
  _dz = uaxis[0].d; _dx = uaxis[1].d; _dy = uaxis[2].d; _dtu = uaxis[3].d;
  _onestp = _nx*_nz*_ny;
  std::vector<axis> azxy;
  azxy.push_back(uaxis[0]); azxy.push_back(uaxis[1]); azxy.push_back(uaxis[2]);

  _lapop = lapop;

  /* Get velocity and frequency info */
  _fmax = fmax; _vmax = vel->my_max(); _vmin = find_vmin(vel);

  /* Check spatial samplings */
  if(!check_dxdz(lapop)) {
    seperr("Spatial samplings in 3D laplacian operator are not equivalent to "
        "afd3d spatial samplings. Please check again.\n");
  }

  /* Check for stability */
  if(!check_cfl(_vmax,lapop)) {
    fprintf(stderr,"CFL is: %f\n",get_cfl());
    seperr("afd3d: Error, please change your samplings. Your propagation will be"
        " unstable.\n");
  }

  /* Check for dispersion */
  if(!check_ptsperlength(_vmin, _fmax,lapop)) {
    fprintf(stderr,"PPL is: %f\n",get_ptsperlength());
    seperr("afd3d: Error, please change your samplings. Your propagation will be"
        " dispersive.\n");
  }

  /* Precompute velocity dt^2 coeffiecient */
  _v2dt2 = (hypercube_float*)vel->clone_vec();
  _v2dt2->mult(_v2dt2);
  _v2dt2->scale(_dtu * _dtu);

  /* Allocate memory for wavefield slices*/
  _pre   = new hypercube_float(azxy, true);
  _cur   = new hypercube_float(azxy, true);
  _nex   = new hypercube_float(azxy, true);
  _lap   = new hypercube_float(azxy, true);
  _tmp   = new hypercube_float(azxy, true);
  _sou   = new hypercube_float(azxy, true);
}

bool afd3d::forward(bool add, my_vector *mod, my_vector *dat) {
  if(add && dat) dat->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  /* Initialize temporary cubes */
  _pre->zero(); _cur->zero(); _nex->zero(); _tmp->zero(); _sou->zero();

  /* it = -1 (u'(0) = 0) */
  memcpy(d->vals,_pre->vals,sizeof(float)*_onestp);
  /* it = 0 (u(0) = 0) */
  memcpy(&d->vals[_onestp],_pre->vals,sizeof(float)*_onestp);

  /* Time Loop */
  for(int it = 1; it < _ntu-1; ++it) {
    fprintf(stderr,"\b\b\b\b\b\b\bit=%d",it);
    /* Forward march */
    fmarch(_pre,_cur,_nex);
    /* Add source */
    memcpy(_sou->vals,&m->vals[(it)*_onestp],sizeof(float)*_onestp);
    _nex->add(_sou);
    /* Save to wavefield */
    memcpy(&d->vals[it*_onestp],_nex->vals,sizeof(float)*_onestp);
    /* Update time step */
    _tmp = _pre; _pre = _cur; _cur = _nex; _nex = _tmp;
  }
  return true;
}

bool afd3d::adjoint(bool add, my_vector *mod, my_vector *dat) {
  return true;
}

void afd3d::fmarch(hypercube_float *pre, hypercube_float *cur, hypercube_float *nex) {
  _lapop->forward(false,cur,_lap);
  for(int k = 0; k < _onestp; ++k) {
    nex->vals[k] = _v2dt2->vals[k]*_lap->vals[k] + 2*cur->vals[k] - pre->vals[k];
  }
}

float afd3d::find_vmin(hypercube_float *vel) {
  int shift = _lapop->get_myorder()/2;
  float vmin = vel->vals[shift];

  for(int iy = shift; iy < _ny-shift; ++iy) {
    for(int ix = shift; ix < _nx-shift; ++ix) {
      for(int iz = shift; iz < _nz-shift; ++iz) {
        if(vmin < vel->vals[iy*_nx*_nz + ix*_nz + iz])
          vmin = vel->vals[iy*_nx*_nz + ix*_nz + iz];
      }
    }
  }
  return vmin;
}

bool afd3d::check_dxdz(laplacian3d *lapop) {
  float idx2 = 1/(_dx*_dx); float idz2 = 1/(_dz*_dz); float idy2 = 1/(_dy*_dy);
  if(lapop->_idx2 != idx2 && lapop->_idz2 != idz2 && lapop->_idy2 != idy2) {
    return false;
  } else {
    return true;
  }
}

bool afd3d::check_cfl(float vmax, laplacian3d* lapop) {
  /* Compute condition based on odd laplacian coefficients */
  float sum = 0.0;
  std::vector<double> coefs = lapop->get_mycoefs();
  for(int i = 0; i < (int)coefs.size()/2 +1; ++i) {sum += coefs[2*i-1]; }
  sum *= 3; _CFL = sqrt(1/sum);
  /* Compute my CFL */
  _mycfl = (_dtu * vmax)/(std::min({_dx,_dz,_dy}));
  if(_mycfl < _CFL) {
    return true;
  } else {
    return false;
  }
}

bool afd3d::check_ptsperlength(float vmin, float fmax, laplacian3d* lapop) {
  _ppl = vmin/(fmax*std::max({_dx,_dz,_dy}));
  if(_lapop->get_myorder() == 4) {
    if(_ppl < 4) {
      return false;
    } else {
      return true;
    }
  }
  //TODO: How can this be derived?
  else if(_lapop->get_myorder() == 10) {
    if(_ppl < 3) {
      return false;
    }
    else {
      return true;
    }
  }
  else{
    seperr("ACFDprop2d has not been implemented for stencil orders besides 4 and 10.\n");
    return false;
  }
}
std::pair<float,float> afd3d::get_cfl() {
  check_cfl(_vmax,_lapop);
  return std::pair<float,float>(_mycfl,_CFL);
}

float afd3d::get_ptsperlength() {
  check_ptsperlength(_vmin,_fmax,_lapop);
  return _ppl;
}


