/**
 * Generates weights for absorbing boundaries
 * in finite difference wave propagation
 * @author: Joseph Jennings
 * @version: 2016.11.27
 */

#include "sep_params.h"
#include "hypercube_float.h"
#include "simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);

  int ord = pars.get_int("ord",4);

  /* Get velocity */
  sregf velin = sregf("in");
  int nx   = velin.get_axis(1).n;
  int nz   = velin.get_axis(2).n;
  float dx = velin.get_axis(1).d;
  float dz = velin.get_axis(2).d;

  hypercube_float *vel = new hypercube_float(velin.get_axes(), true);
  hypercube_float *vwt = new hypercube_float(velin.get_axes(), true);
  hypercube_float *wgt = new hypercube_float(velin.get_axes(), true);

  velin.read_all("in",vel);
  vwt->zero(); wgt->zero();

  /* Get command line parameters */
  int bz1 = pars.get_int("bz1",0);
  int bx1 = pars.get_int("bx1",0);
  int bz2 = pars.get_int("bz2",nz);
  int bx2 = pars.get_int("bx2",nx);
  float dtu = pars.get_float("dtu",0.001);

  int shift = ord/2;

  /* First iz boundary */
  for(int ix = shift; ix < nx - shift; ++ix) {
    for(int iz = shift; iz < bz1 + shift; ++iz) {
      vwt->vals[ix*nz + iz]  = vel->vals[ix*nz + iz]*dtu/dz;
      wgt->vals[ix*nz + iz] += (static_cast<float>(bz1-(iz-shift))/static_cast<float>(bz1)) * (0.15)*vwt->vals[ix*nz + iz];
    }
  }

  /* Second iz boundary */
  for(int ix = shift; ix < nx - shift; ++ix) {
    for(int iz = nz-bz2-shift; iz < nz - shift; ++iz) {
      vwt->vals[ix*nz + iz]  = vel->vals[ix*nz + iz]*dtu/dz;
      wgt->vals[ix*nz + iz] += (static_cast<float>(iz-nz+bz2+1+shift)/static_cast<float>(bz2)) * (0.15)*vwt->vals[ix*nz + iz];
    }
  }

  /* First ix boundary */
  for(int ix = shift; ix < bx1+shift; ++ix) {
    for(int iz = shift; iz < nz-shift; ++iz) {
      vwt->vals[ix*nz + iz]  = vel->vals[ix*nz + iz]*dtu/dx;
      wgt->vals[ix*nz + iz] += (static_cast<float>(bx1-(ix-shift))/static_cast<float>(bx1)) * (0.15)*vwt->vals[ix*nz + iz];
    }
  }

  /* Second ix boundary */
  for(int ix = nx-bx2-shift; ix < nx-shift; ++ix) {
    for(int iz = shift; iz < nz-shift; ++iz) {
      vwt->vals[ix*nz + iz]  = vel->vals[ix*nz + iz]*dtu/dx;
      wgt->vals[ix*nz + iz] += (static_cast<float>(ix-nx+bx2+1+shift)/static_cast<float>(bx2)) * (0.15)*vwt->vals[ix*nz + iz];
    }
  }

  /* Write out weights */
  sregf wtout = sregf(wgt->get_axes());
  /* Write out velocity weight */
  wtout.write_description("vwt");
  wtout.write_all("vwt",vwt);
  /* Write out weight */
  wtout.write_description("out");
  wtout.write_all("out",wgt);

  /* Clean up */
  delete vel; delete vwt; delete wgt;

}
