/**
 * Migrates shot gathers
 * given data, a background velocity,
 * and an input source wavelet
 * @author: Joseph Jennings
 * @version: 2017.01.22
 */

#include "sep_params.h"
#include "../main/afd2dfst.h"
#include "../main/bscat.h"
#include "../main/sourcescale.h"
#include "../main/scndder.h"
#include "../main/rrec.h"
#include "../main/zeropad.h"
#include "../main/laplacian10th.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  /* Source axes */
  float srcz = pars.get_float("srcz",50.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);

  /* Output file for wavefields */
  std::string bwfname = pars.get_string("bwfldsout","");
  std::string awfname = pars.get_string("awfldsout","");

  /* Input files */
  sregf datin = sregf("in");
  sregf srcin = sregf("src");
  sregf velin = sregf("vel");
  sregf wfout;

  if(!datin.get_axis(1).same_axis(srcin.get_axis(1)))
    seperr("MigrateShots: input data and source wavelet must have same time axis.\n");

  /* Velocity axes */
  int   nz  = velin.get_axis(1).n;
  float dz  = velin.get_axis(1).d;
  int   nx  = velin.get_axis(2).n;
  float dx  = velin.get_axis(2).d;
  /* Data axes */
  int   nt  = datin.get_axis(1).n;
  int   nrx = datin.get_axis(2).n;
  int   nsx = datin.get_axis(3).n;
  float osx = datin.get_axis(3).o;
  float dsx = datin.get_axis(3).d;

  /* Build axes */
  std::vector<axis> raxes;
  raxes.push_back(datin.get_axis(1));
  raxes.push_back(datin.get_axis(2));

  std::vector<axis>paxes;
  paxes.push_back(axis(nz,0.0,dz));
  paxes.push_back(axis(nx,0.0,dx));
  paxes.push_back(srcin.get_axis(1));

  /* Build hypercubes */
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *prc = new hypercube_float(paxes,true);
  hypercube_float *bck = new hypercube_float(paxes,true);
  hypercube_float *ssc = new hypercube_float(paxes,true);
  hypercube_float *sht = new hypercube_float(raxes,true);
  hypercube_float *rec = new hypercube_float(raxes,true);
  hypercube_float *dat = new hypercube_float(datin.get_axes(),true);
  hypercube_float *img = new hypercube_float(velin.get_axes(),true);

  /* Read in input files */
  datin.read_all("in",dat);
  srcin.read_all("src",src);
  velin.read_all("vel",vel);
  if(!bwfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(bwfname);
  }
  if(!awfname.empty() && !bwfname.empty()) {
    wfout.write_description(awfname);
  } else if(!awfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(awfname);
  }

  /* Construct operators */
  sourcescale ss    = sourcescale(dtu,vel);
  laplacian10th *lp = new laplacian10th(paxes);
  afd2dfst fd       = afd2dfst(paxes,vel,lp,dtu,fmax);
  rrec rr           = rrec(datin.get_axis(2),recz);
  scndder d2        = scndder(datin.get_axis(1));
  bscat   im        = bscat(prc);

  /* Loop over shot */
  for(int is = 0; is < nsx; ++is) {
    /* Compute background wavefield */
    float xs = osx + dsx*is;
    zeropad zp = zeropad(static_cast<int>(xs),static_cast<int>(srcz));
    zp.forward(false,src,prc);
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    rr.adjoint(false,rec,prc);
    /* If user specifies, write out background wavefield */
    if(!bwfname.empty()) {
      wfout.write_all(bwfname,prc);
      wfout.write_axis(bwfname,4,axis(is+1,0,1.0,"shot"));
    }
    /* Compute data perturbation */
    memcpy(sht->vals,&dat->vals[nt*nrx*is],sizeof(float)*nt*nrx);
    rec->scale_add(1.0,sht,-1.0);
    /* Back propagate perturbation */
    rr.forward(false,rec,ssc);
    fd.adjoint(false,bck,ssc);
    /* If user specifies, write out adjoint wavefield */
    if(!awfname.empty()) {
      wfout.write_all(awfname,bck);
      wfout.write_axis(awfname,4,axis(is+1,0,1.0,"shot"));
    }
    ss.adjoint(false,ssc,bck);
    /* Imaging condition */
    d2.adjoint(false,bck,ssc);
    im.adjoint(true,img,bck);
  }

  /* Write Image */
  sregf imgout = sregf(img->get_axes());
  imgout.write_description("out");
  imgout.write_all("out",img);
  imgout.write_axis("out",3,axis(1,0.0,1.0));

  /* Clean up */
  delete vel; delete src; delete dat;
  delete lp; delete img; delete bck; delete sht;
  delete rec;
}
