/**
 * Model a 2D survey given
 * an input source, velocity model and
 * acquisition parameters
 * @author: Joseph Jennings
 * @version: 2017.01.21
 */

#include "sep_params.h"
#include "../main/zeropad.h"
#include "../main/sourcescale.h"
#include "../main/afd2dfst.h"
#include "../main/laplacian10th.h"
#include "../main/rrec.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  /* Source axes */
  int nsx    = pars.get_int  ("nsx",  1);
  float osx  = pars.get_float("osx",0.0);
  float dsx  = pars.get_float("dsx",1.0);
  float srcz = pars.get_float("srcz",50.0);
  /* Receiver axes */
  int nrx    = pars.get_int  ("nrx",  1);
  float orx  = pars.get_float("orx",0.0);
  float drx  = pars.get_float("drx",1.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);

  /* Output file for wavefields */
  std::string wfname = pars.get_string("wfldsout","");

  /* Input files */
  sregf srcin = sregf("in");
  sregf velin = sregf("vel");
  sregf wfout;

  int   nz = velin.get_axis(1).n;
  float dz = velin.get_axis(1).d;
  int   nx = velin.get_axis(2).n;
  float dx = velin.get_axis(2).d;
  int   nt = srcin.get_axis(1).n;

  std::vector<axis> daxes,raxes;
  daxes.push_back(srcin.get_axis(1)); raxes.push_back(srcin.get_axis(1));
  daxes.push_back(axis(nrx,orx,drx)); raxes.push_back(axis(nrx,orx,drx));
  daxes.push_back(axis(nsx,0.0,1.0));

  std::vector<axis>paxes;
  paxes.push_back(axis(nz,0.0,dz));
  paxes.push_back(axis(nx,0.0,dx));
  paxes.push_back(srcin.get_axis(1));

  /* Build hypercubes */
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *prc = new hypercube_float(paxes,true);
  hypercube_float *ssc = new hypercube_float(paxes,true);
  hypercube_float *rec = new hypercube_float(raxes,true);
  hypercube_float *dat = new hypercube_float(daxes,true);

  /* Read in input files */
  srcin.read_all("in",src);
  velin.read_all("vel",vel);
  if(!wfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(wfname);
  }

  /* Construct operators */
  sourcescale ss = sourcescale(dtu,vel);
  laplacian10th *lp = new laplacian10th(paxes);
  afd2dfst fd = afd2dfst(paxes,vel,lp,dtu,fmax);
  rrec rr = rrec(axis(nrx,orx,drx),recz);

  /* Loop over shot */
  for(int is = 0; is < nsx; ++is) {
    float xs = osx + dsx*is;
    zeropad zp = zeropad(static_cast<int>(xs),static_cast<int>(srcz));
    zp.forward(false,src,prc);
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out wavefield */
    if(!wfname.empty()) {
      wfout.write_all(wfname,prc);
      wfout.write_axis(wfname,4,axis(is+1,0,1.0,"shot"));
    }
    rr.adjoint(false,rec,prc);
    memcpy(&dat->vals[is*nt*nrx],rec->vals,sizeof(float)*nt*nrx);
  }

  /* Write out data */
  sregf datout = sregf(dat->get_axes());
  datout.write_description("out");
  datout.write_all("out",dat);

  /* Clean up */
  delete vel; delete src; delete prc; delete dat; delete rec;
  delete lp;
}
