/**
 * A program to test the constant density,
 * 2D FD wave propagation operator
 * @author: Joseph Jennings
 * @version: 2017.01.14
 */

#include "sep_params.h"
#include "../main/acfdprop2d.h"
#include "../main/afd2dfst.h"
#include "../main/laplacian4th.h"
#include "../main/laplacian10th.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);

  int adj = pars.get_int("adj",0);
  int ord = pars.get_int("ord",4);
  float fmax = pars.get_float("fmax",50.0);
  std::string wtname = pars.get_string("wgt","");
  std::string vwname = pars.get_string("vwt","");
  std::string mode   = pars.get_string("mode","slow");
  int bz = 0; int bx = 0; float alpha = 0.0;
  float otu = 0.0; float dtu = 0.0;
  if(!mode.compare("fast")) {
    bz = pars.get_int("bz",0);
    bx = pars.get_int("bx",0);
    alpha = pars.get_float("alfa",0.0);
    otu = pars.get_float("otu",0.0);
    dtu = pars.get_float("dtu",0.001);
  }

  sregf srcin; sregf wfdin;
  if(adj) {
    wfdin = sregf("in");
  } else {
    srcin = sregf("in");
  }
  sregf velin = sregf("vel");
  sregf wgtin; sregf vwtin;
  bool bndrs = false;
  if(!wtname.empty() && !vwname.empty()) {
    wgtin = sregf("wgt");
    vwtin = sregf("vwt");
    bndrs = true;
  } else if(!wtname.empty() || !vwname.empty()) {
    seperr("Must provide both a weight function and velocity weight function"
        "for absorbing boundaries.\n");
  }

  std::vector<int>nbs; nbs.resize(4);
  if(bndrs) {
    nbs[0]  = pars.get_int("bz1",0);
    nbs[2]  = pars.get_int("bz2",wgtin.get_axis(1).n);
    nbs[1]  = pars.get_int("bx1",0);
    nbs[3]  = pars.get_int("bx2",wgtin.get_axis(2).n);
  }

  hypercube_float *src = nullptr; hypercube_float *wfd = nullptr;
  hypercube_float *vel = new hypercube_float(velin.get_axes(), true);
  hypercube_float *wgt = nullptr; hypercube_float *vwt = nullptr;
  if(bndrs) {
    wgt = new hypercube_float(wgtin.get_axes(), true);
    vwt = new hypercube_float(vwtin.get_axes(), true);
  }

  if(adj) {
    /* Build wavefield axis */
    wfd = new hypercube_float(wfdin.get_axes(), true);
    /* Build source axis */
    std::vector<axis> saxis;
    saxis.push_back(axis(vel->get_axis(1).n,0.0,1.0));
    saxis.push_back(axis(vel->get_axis(2).n,0.0,1.0));
    saxis.push_back(wfdin.get_axis(3));
    //saxis.push_back(axis(ntu,otu,dtu));
    src = new hypercube_float(saxis, true);
  } else {
    src = new hypercube_float(srcin.get_axes(), true);
    std::vector<axis> uaxis;
    uaxis.push_back(vel->get_axis(1)); uaxis.push_back(vel->get_axis(2));
    if(!mode.compare("fast")) {
      uaxis.push_back(axis(srcin.get_axis(3).n,otu,dtu));
    } else if(!mode.compare("slow")) {
      uaxis.push_back(src->get_axis(3));
    } else {
      seperr("Error: mode %s not supported. Please try 'fast' or 'slow'.\n",mode.c_str());
    }
    wfd = new hypercube_float(uaxis, true);
  }

  velin.read_all("vel",vel);
  if(bndrs) {
    wgtin.read_all("wgt",wgt);
    vwtin.read_all("vwt",vwt);
  }

  laplacian *lap = nullptr;
  if(ord == 4) {
    lap = new laplacian4th(wfd->get_axes());
  } else if(ord == 10) {
    lap = new laplacian10th(wfd->get_axes());
  } else {
    seperr("Please either choose 4th or 10th order. Order %d has not yet been implemented.\n",ord);
  }

  acfdprop2d *prp = nullptr; afd2dfst *fprp = nullptr;
  if(bndrs) {
    prp = new acfdprop2d(wfd->get_axes(),vel,wgt,vwt,nbs,lap,fmax);
    fprintf(stderr,"CFL=%f\n",prp->get_cfl());
  }
  else {
    if(!mode.compare("slow")) {
      prp = new acfdprop2d(wfd->get_axes(),vel,lap,fmax);
      fprintf(stderr,"CFL=%f\n",prp->get_cfl());
    }
    else if(!mode.compare("fast")) {
      if(!bz) {
        fprp = new afd2dfst(src->get_axes(), vel, lap, dtu, fmax);
        fprintf(stderr,"mycfl=%f lapcfl=%f\n",fprp->get_cfl().first,fprp->get_cfl().second);
      } else {
        if(adj) {
          std::vector<axis> csaxis;
          csaxis.push_back(wfd->get_axis(1)); csaxis.push_back(wfd->get_axis(2));
          csaxis.push_back(axis(wfdin.get_axis(3).n,otu,wfdin.get_axis(3).d));
          //fprp = new afd2dfst(csaxis, vel, lap, fmax, dtu, bz, bx, alpha);
          fprp = new afd2dfst(wfdin.get_axes(), vel, lap, dtu, fmax, bz, bx, alpha);
        }
        else {
          fprp = new afd2dfst(src->get_axes(), vel, lap, dtu, fmax, bz, bx, alpha);
          fprintf(stderr,"mycfl=%f lapcfl=%f\n",fprp->get_cfl().first,fprp->get_cfl().second);
        }
      }
    }
  }

  if(adj) {
    wfdin.read_all("in",wfd);
    if(!mode.compare("slow")) {
      prp->adjoint(false,src,wfd);
    }
    else if(!mode.compare("fast")) {
      fprp->adjoint(false,src,wfd);
    }
    sregf sout = sregf(src->get_axes());
    sout.write_description("out");
    sout.write_all("out",src);
  }
  else {
    srcin.read_all("in",src);
    if(!mode.compare("slow")) {
      prp->forward(false,src,wfd);
    } else if(!mode.compare("fast")) {
      fprp->forward(false,src,wfd);
    }
    std::vector<axis> mix;
    mix.push_back(wfd->get_axis(1));
    mix.push_back(wfd->get_axis(2));
    mix.push_back(axis(wfd->get_axis(3).n,wfd->get_axis(3).o,srcin.get_axis(3).d));
    sregf wout;
    if(!mode.compare("slow")) {
      wout = sregf(wfd->get_axes());
    } else if(!mode.compare("fast")) {
      wout = sregf(mix);
    }
    wout.write_description("out");
    wout.write_all("out",wfd);
  }

  //delete prp; delete fprp;
  delete src; delete wfd; delete vel; delete lap;
}
