/**
 * A main program to test frequency
 * filtering using FFTW
 * @author: Joseph Jennings
 * @version: 2016.12.10
 */

#include "sep_params.h"
#include "sregf.h"
#include "sregc.h"
#include <complex>
#include <fftw3.h>
#include "hypercube_float.h"
#include "hypercube_complex.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  /* Get command line parameters */
  sregf sigin = sregf("in");
  std::vector<float> fdef  = {2.0,10.0,20.0,22.0};
  std::vector<float> frqs = pars.get_vfloats("freqs",4,fdef);
  float fnot = pars.get_float("fnot",9.0);
  std::string mode = pars.get_string("mode","bnd");
  if(frqs.size() < 4) {
    for(int f = frqs.size(); f < 4; ++f)  frqs.push_back(fdef[f]);
  }

  /* Input signal*/
  hypercube_float *sig = new hypercube_float(sigin.get_axes(), true);
  sigin.read_all("in",sig);

  /* Time axis */
  int nt   = sigin.get_axis(1).n;
  float ot = sigin.get_axis(1).o;
  float dt = sigin.get_axis(1).d;

  /* Frequency axis */
  int nw = nt/2;
  float dw = 1/(nt*dt);

  /* Output frequency axis */
  std::vector<axis> af; af.push_back(axis(nt/2,ot,1/(nt*dt)));

  std::complex<double> *in   = new std::complex<double>[nt];
  std::complex<double> *in2  = new std::complex<double>[nt];
  std::complex<double> *out  = new std::complex<double>[nt];
  std::complex<double> *out2 = new std::complex<double>[nt];
  std::complex<double>* flt  = new std::complex<double>[nt];

  for(int it = 0; it < nt; ++it) {
    in[it] = std::complex<double>(static_cast<double>(sig->vals[it]),0.0);
  }

  /* Compute the FFT */
  fftw_plan p = fftw_plan_dft_1d(nt,reinterpret_cast<fftw_complex*>(in),
      reinterpret_cast<fftw_complex*>(out),FFTW_FORWARD,FFTW_ESTIMATE);
  fftw_execute(p);

  /* Notch filter */
  if(!mode.compare("not")) {
    for(int iw = 0; iw < nw; ++iw) {
      flt[iw] = std::complex<double>(1.0,0.0);
      out2[iw] = out[iw]*flt[iw];
      float freq = iw*dw;
      if(freq == fnot) {
        flt[iw] = std::complex<double>(0.0,0.0);
        out2[iw]      = out[iw]*flt[iw];
        out2[nt-1-iw] = out[nt-1-iw]*flt[iw];
      }
    }
  }
  /* Bandpass */
  else if(!mode.compare("bnd")) {
    std::complex<double> coef; double coefr; double coefi = 0.0;
    for(int iw = 0; iw < nw; ++iw) {
      float freq = dw*iw;
      if(freq < frqs[0]) {
        coefr         = 0.0;
        coef          = std::complex<double>(coefr,coefi);
        flt[iw     ]  = coef;
        flt[nt-iw-1]  = coef;
        out2[iw]      = out[iw]*coef;
        out2[nt-iw-1] = out[nt-iw-1]*coef;
      } else if(freq >= frqs[0] && freq < frqs[1]) {
        coefr         = cos(M_PI/2.0 * (frqs[1]-freq)/(frqs[1]-frqs[0]))*cos(M_PI/2.0 * (frqs[1]-freq)/(frqs[1]-frqs[0]));
        coef          = std::complex<double>(coefr,coefi);
        flt[iw     ]  = coef;
        flt[nt-iw-1]  = coef;
        out2[iw]      = out[iw]*coef;
        out2[nt-iw-1] = out[nt-iw-1]*coef;
      } else if(freq >= frqs[1] && freq < frqs[2]) {
        coefr         = 1.0;
        coef          = std::complex<double>(coefr,coefi);
        flt[iw     ]  = coef;
        flt[nt-iw-1]  = coef;
        out2[iw]      = out[iw]*coef;
        out2[nt-iw-1] = out[nt-iw-1]*coef;
      } else if(freq >= frqs[2] && freq < frqs[3]){
        coefr         = cos(M_PI/2.0 * (freq-frqs[2])/(frqs[3]-frqs[2]))*cos(M_PI/2.0 * (freq-frqs[2])/(frqs[3]-frqs[2]));
        coef          = std::complex<double>(coefr,coefi);
        flt[iw     ]  = coef;
        flt[nt-iw-1]  = coef;
        out2[iw]      = out[iw]*coef;
        out2[nt-iw-1] = out[nt-iw-1]*coef;
      } else {
        coefr         = 0.0;
        coef          = std::complex<double>(coefr,coefi);
        flt[iw     ]  = coef;
        flt[nt-iw-1]  = coef;
        out2[iw]      = out[iw]*coef;
        out2[nt-iw-1] = out[nt-iw-1]*coef;
      }
    }
  }

  /* Inverse FFT */
  fftw_plan q = fftw_plan_dft_1d(nt,reinterpret_cast<fftw_complex*>(out2),
      reinterpret_cast<fftw_complex*>(in2),
      FFTW_BACKWARD,FFTW_ESTIMATE);

  fftw_execute(q);

  /* Build hypercubes */
  hypercube_float *fsig    = new hypercube_float(sig->get_axes(),true);
  for(int it = 0; it < nt; ++it) fsig->vals[it] = in2[it].real();
  hypercube_complex *ftin  = new hypercube_complex(af,out);
  hypercube_complex *spctm = new hypercube_complex(af,out2);
  hypercube_complex *fltsp = new hypercube_complex(af,flt);

  /* Filtered signal */
  fsig->scale(static_cast<double>(1.0/nt));
  sigin.write_description("out");
  sigin.write_all("out",fsig);

  /* Filtered spectrum */
  sregc spcout = sregc(af);
  spcout.write_description("spec");
  spcout.write_all("spec",spctm);

  /* Fourier transform of input */
  spcout.write_description("ftin");
  spcout.write_all("ftin",ftin);

  /* Frequency domain filter */
  sregc fltout = sregc(af);
  fltout.write_description("flt");
  fltout.write_all("flt",fltsp);

  delete sig; delete spctm; delete fltsp;
  delete flt; delete in; delete in2;
  delete out; delete ftin;
}
