/**
 * A simple 1D example for adding two
 * signals of different sampling rates
 * and lengths
 * @author: Joseph Jennings
 * @version: 2017.01.05
 */

#include "sep_params.h"
#include "sregf.h"
#include "hypercube_float.h"
#include "../main/lininterp.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  /* Input IO */
  sep_params pars = sep_params(argc, argv);
  int adj = pars.get_int("adj",0);

  sregf crsin = sregf("in");
  sregf finin = sregf("fine");

  /* Coarse axis */
  int nts   = crsin.get_axis(1).n;
  float dts = crsin.get_axis(1).d;

  /* Fine axis */
  int   ntu = finin.get_axis(1).n;
  float dtu = finin.get_axis(1).d;

  /* Read in data */
  hypercube_float *crs = new hypercube_float(crsin.get_axes(), true);
  hypercube_float *fin = new hypercube_float(finin.get_axes(), true);
  hypercube_float *frp = new hypercube_float(finin.get_axes(), true);
  hypercube_float *crp = new hypercube_float(crsin.get_axes(), true);
  hypercube_float *cmp = new hypercube_float(crsin.get_axes(), true);

  crsin.read_all("in",crs);
  finin.read_all("fine",fin);

  /* Precompute interpolation coefficients */
  std::vector<double> ls1, ls2, lr1, lr2; int sample = static_cast<int>(dts/dtu);
  for(int is = 0; is < sample; ++is) {
    ls1.push_back(static_cast<double>((sample-is))/static_cast<double>(sample));
    lr1.push_back(static_cast<double>((sample-is-1))/static_cast<double>(sample));
    ls2.push_back(1.0-ls1[is]);
    lr2.push_back(1.0-lr1[is]);
    //ls1[is] /= sample; ls2[is] /= sample;
    fprintf(stderr,"is=%d ls1=%f ls2=%f\n",is,ls1[is],ls2[is]);
    fprintf(stderr,"is=%d lr1=%f lr2=%f\n",is,lr1[is],lr2[is]);
  }

  if(adj) {
    /* Reverse loop over coarsely sampled signal (data) */
    float tval1 = 0.0; float tval2 = 0.0;
    int k = ntu-2*sample;  //Initial condition
    for(int it = nts-2; it >= 1; --it) {
      /* Extract two source slices */
      float val1 = crs->vals[it];
      float val2 = crs->vals[it-1];
      /* Compute one coarse step */
      for(int is = 0; is < sample; ++is) {
        --k; // This is equivalent to putting the forward march for wave propagation.
        /* Forward interpolate source and add */
        frp->vals[k] = fin->vals[k] + lr1[is]*val1 + lr2[is]*val2;
        //fprintf(stderr,"it=%d k=%d frp=%f fx=%f gx=%f m=%f mp1=%f\n",it,k,frp->vals[k],ls2[is],ls1[is],val1,val2,fin->vals[k]);
        /* Adjoint interpolate the wavefield */
        tval1 += frp->vals[k]*lr1[is];
        tval2 += frp->vals[k]*lr2[is];
        //fprintf(stderr,"it=%d k=%d frp=%f fx=%f gx=%f m=%f mp1=%f\n",it,k,frp->vals[k],ls2[is],ls1[is],tval1,tval2);
      }
      crp->vals[it]   = tval1;
      crp->vals[it-1] = tval2;
      tval1 = tval2;
      tval2 = 0.0;
    }
  }
  else {
    /* Forward Loop over coarsely sampled signal (source) */
    float tval1 = 0.0; float tval2 = 0.0;
    int k = -1;  //Initial condition
    for(int it = 1; it < nts-1; ++it) {
      /* Extract two source slices */
      float val1 = crs->vals[it-1];
      float val2 = crs->vals[it];
      /* Compute one coarse step */
      for(int is = 0; is < sample; ++is) {
        ++k; // This is equivalent to putting the forward march for wave propagation.
        /* Forward interpolate source and add */
        frp->vals[k] = fin->vals[k] + ls1[is]*val1 + ls2[is]*val2;
        //frp->vals[k] = ls1[is]*val1 + ls2[is]*val2;
        fprintf(stderr,"it=%d k=%d frp=%f fx=%f gx=%f m=%f mp1=%f fin=%f\n",it,k,frp->vals[k],ls2[is],ls1[is],val1,val2,fin->vals[k]);
        /* Adjoint interpolate the wavefield */
        tval1 += frp->vals[k]*ls1[is];
        tval2 += frp->vals[k]*ls2[is];
        //fprintf(stderr,"it=%d k=%d frp=%f fx=%f gx=%f m=%f mp1=%f\n",it,k,frp->vals[k],ls2[is],ls1[is],tval1,tval2);
      }
      crp->vals[it-1]   = tval1;
      crp->vals[it] = tval2;
      tval1 = tval2;
      tval2 = 0.0;
    }
  }
  /* Adjoint linear interp for comparison */
  lininterp li = lininterp(1);

  /* Write out interpolated signal */
  crsin.write_description("out");
  crsin.write_all("out",crp);

  /* Write out comparison signal */
  sregf cmpout = sregf(cmp->get_axes());

  /* Write out adjoint interpolated frp */
  li.adjoint(false,cmp,frp);
  cmpout.write_description("afinout");
  cmpout.write_all("afinout",cmp);

  /* Write out fine interpolated signal */
  finin.write_description("finout");
  finin.write_all("finout",frp);

  /* Clean up */
  delete crs; delete fin; delete crp; delete frp;
}
