/**
 * Tests the ISPC implementation
 * of the 10th-order laplacian
 * @author: Joseph Jennings
 * @version: 2017.04.05
 */

#include <limits.h>
#include "gtest/gtest.h"
#include "../main/laplacian10th.h"
#include "../main/laplacian10thispc.h"
#include "axis.h"
#include "hypercube_float.h"

class laplacianISPCtest: public ::testing::Test {
  public:
    void trunc(int loc1, int loc2, hypercube_float *dat, hypercube_float* pad);
    void pad(int loc1, int loc2, hypercube_float *dat, hypercube_float* pad);
  protected:
    virtual void SetUp(){
    }

    virtual void TearDown() {
    }
};

TEST_F(laplacianISPCtest,fwdtest){
  int nx = 100; int nz = 100;
  int p = 10;
  /* Build axes */
  axis ax  = axis(nx  , 0.0, 1.0);
  axis az  = axis(nz  , 0.0, 1.0);
  axis axp = axis(nx+p, 0.0, 1.0);
  axis azp = axis(nz+p, 0.0, 1.0);
  std::vector<axis> axes, paxes;
  axes.push_back(az); axes.push_back(ax);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Build hypercubes */
  hypercube_float *rm = new hypercube_float(axes, true);
  hypercube_float *rd = new hypercube_float(axes, true);
  hypercube_float *m  = new hypercube_float(paxes, true);
  hypercube_float *d  = new hypercube_float(paxes, true);
  hypercube_float *ds = new hypercube_float(paxes, true);
  rm->random(); rd->random();
  pad(5,5,rm,m); pad(5,5,rd,d);
  laplacian10th lap = laplacian10th(paxes);
  laplacian10thispc lapis = laplacian10thispc(paxes);
  lap.forward(false,m,ds);
  lapis.forward(false,m,d);
  ASSERT_NEAR(d->sum(),ds->sum(),0.00001);
  delete m;
  delete d; delete ds;
}

TEST_F(laplacianISPCtest,adjtest){
  int nx = 100; int nz = 100;
  int p = 10;
  /* Build axes */
  axis ax  = axis(nx  , 0.0, 1.0);
  axis az  = axis(nz  , 0.0, 1.0);
  axis axp = axis(nx+p, 0.0, 1.0);
  axis azp = axis(nz+p, 0.0, 1.0);
  std::vector<axis> axes, paxes;
  axes.push_back(az); axes.push_back(ax);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Build hypercubes */
  hypercube_float *rm = new hypercube_float(axes, true);
  hypercube_float *rd = new hypercube_float(axes, true);
  hypercube_float *m  = new hypercube_float(paxes, true);
  hypercube_float *ms = new hypercube_float(paxes, true);
  hypercube_float *d  = new hypercube_float(paxes, true);
  rm->random(); rd->random();
  pad(5,5,rm,m); pad(5,5,rd,d);
  laplacian10th lap = laplacian10th(paxes);
  laplacian10thispc lapis = laplacian10thispc(paxes);
  lap.adjoint(false,ms,d);
  lapis.adjoint(false,m,d);
  ASSERT_NEAR(m->sum(),ms->sum(),0.0001);
  delete m;
  delete d; delete ms;
}

TEST_F(laplacianISPCtest,dptest) {
  int nx = 100; int nz = 100;
  int p = 10;
  /* Build axes */
  axis ax  = axis(nx  , 0.0, 1.0);
  axis az  = axis(nz  , 0.0, 1.0);
  axis axp = axis(nx+p, 0.0, 1.0);
  axis azp = axis(nz+p, 0.0, 1.0);
  std::vector<axis> axes, paxes;
  axes.push_back(az); axes.push_back(ax);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Build hypercubes */
  hypercube_float *rm = new hypercube_float(axes, true);
  hypercube_float *rd = new hypercube_float(axes, true);
  hypercube_float *m  = new hypercube_float(paxes, true);
  hypercube_float *ms = new hypercube_float(paxes, true);
  hypercube_float *d  = new hypercube_float(paxes, true);
  hypercube_float *ds = new hypercube_float(paxes, true);
  rm->random(); rd->random();
  pad(5,5,rm,m); pad(5,5,rd,d);
  laplacian10thispc lapis = laplacian10thispc(paxes);
  /* Add == false */
  lapis.forward(false,m,ds);
  lapis.adjoint(false,ms,d);
  double dotd = d->dot(ds);
  double dotm = m->dot(ms);
  ASSERT_NEAR(dotd,dotm,0.00001);
  delete m; delete ds;
  delete d; delete ms;
}

/**
 * Pads a 2D image
 * @param loc1 the offset along axis 1
 * @param loc2 the offset along axis 1
 * @param dat  the input data
 * @param pad  the padded result
 */
void laplacianISPCtest::pad(int loc1, int loc2, hypercube_float *dat, hypercube_float* pad) {
  pad->zero();
  int n1p = pad->get_axis(1).n; int n2p = pad->get_axis(2).n;
  int n1  = dat->get_axis(1).n; int n2  = dat->get_axis(2).n;
  for(int i2 = 0; i2 < n2; ++i2) {
    for(int i1 = 0; i1 < n1; ++i1) {
      pad->vals[(i2+loc2)*n1p + i1 + loc1] = dat->vals[i2*n1 + i1];
    }
  }
}

/**
 * Truncates a 2D image (the adjoint of Pad)
 * @param loc1 the offset along axis 1
 * @param loc2 the offset along axis 1
 * @param dat  the truncated result
 * @param pad  the input padded data
 */
void laplacianISPCtest::trunc(int loc1, int loc2, hypercube_float *dat, hypercube_float* pad) {
  int n1p = pad->get_axis(1).n; int n2p = pad->get_axis(2).n;
  int n1  = dat->get_axis(1).n; int n2  = dat->get_axis(2).n;
  for(int i2 = 0; i2 < n2; ++i2) {
    for(int i1 = 0; i1 < n1; ++i1) {
      dat->vals[i2*n1 + i1] = pad->vals[(i2+loc2)*n1p + i1+loc1];
    }
  }
}
